@@ -173,14 +173,15 @@ totem_action_exit (Totem *totem)
 	if (totem == NULL)
 		exit (0);
 
-	if (totem->playlist != NULL)
-		gtk_widget_hide (GTK_WIDGET (totem->playlist));
-
 	if (totem->win != NULL) {
 		gtk_widget_hide (totem->win);
 		display = gtk_widget_get_display (totem->win);
 	}
 
+	if (totem->prefs != NULL) {
+		gtk_widget_hide (totem->prefs);
+	}
+
 #ifndef HAVE_GTK_ONLY
 	totem_gromit_clear (TRUE);
 #endif /* !HAVE_GTK_ONLY */
@@ -203,8 +204,9 @@ totem_action_exit (Totem *totem)
 				bacon_video_widget_get_volume (totem->bvw),
 				NULL);
 
-	if (totem->playlist)
-		gtk_widget_destroy (GTK_WIDGET (totem->playlist));
+	if (totem->gc)
+		g_object_unref (G_OBJECT (totem->gc));
+
 	if (totem->win)
 		gtk_widget_destroy (GTK_WIDGET (totem->win));
 
@@ -222,6 +224,7 @@ totem_action_menu_popup (Totem *totem, guint button)
 			"totem_right_click_menu");
 	gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL,
 			button, gtk_get_current_event_time ());
+	gtk_menu_shell_select_first (GTK_MENU_SHELL (menu), FALSE);
 }
 
 static gboolean
@@ -535,6 +538,13 @@ totem_action_pause (Totem *totem)
 	}
 }
 
+static void
+totem_action_set_cursor (Totem *totem, gboolean state)
+{
+	totem->cursor_shown = state;
+	bacon_video_widget_set_show_cursor (totem->bvw, state);
+}
+
 static gboolean
 window_state_event_cb (GtkWidget *window, GdkEventWindowState *event,
 		Totem *totem)
@@ -547,7 +557,7 @@ window_state_event_cb (GtkWidget *window, GdkEventWindowState *event,
 		totem_action_save_size (totem);
 		update_fullscreen_size (totem);
 		bacon_video_widget_set_fullscreen (totem->bvw, TRUE);
-		bacon_video_widget_set_show_cursor (totem->bvw, FALSE);
+		totem_action_set_cursor (totem, FALSE);
 
 		if (bacon_video_widget_is_playing (totem->bvw) != FALSE)
 			totem_scrsaver_disable (totem->scr);
@@ -559,7 +569,7 @@ window_state_event_cb (GtkWidget *window, GdkEventWindowState *event,
 
 		popup_hide (totem);
 		bacon_video_widget_set_fullscreen (totem->bvw, FALSE);
-		bacon_video_widget_set_show_cursor (totem->bvw, TRUE);
+		totem_action_set_cursor (totem, TRUE);
 
 		totem_scrsaver_enable (totem->scr);
 
@@ -762,8 +772,10 @@ totem_action_set_mrl_with_warning (Totem *totem, const char *mrl,
 		}
 
 		subtitle_uri = totem_uri_get_subtitle_uri (mrl);
+		totem_gdk_window_set_waiting_cursor (totem->win->window);
 		retval = bacon_video_widget_open_with_subtitle (totem->bvw,
 				mrl, subtitle_uri, &err);
+		gdk_window_set_cursor (totem->win->window, NULL);
 		totem->mrl = g_strdup (mrl);
 
 		/* Play/Pause */
@@ -799,13 +811,20 @@ totem_action_set_mrl_with_warning (Totem *totem, const char *mrl,
 			disp = totem_uri_escape_for_display (totem->mrl);
 			msg = g_strdup_printf(_("Totem could not play '%s'."), disp);
 			g_free (disp);
-			totem_action_error (msg, err->message, totem);
+			if (err && err->message) {
+				totem_action_error (msg, err->message, totem);
+			}
+			else {
+				totem_action_error (msg, _("No error message"), totem);
+			}
 			g_free (msg);
 		}
 
 		if (retval == FALSE)
 		{
-			g_error_free (err);
+			if (err) {
+				g_error_free (err);
+			}
 			g_free (totem->mrl);
 			totem->mrl = NULL;
 			play_pause_set_label (totem, STATE_STOPPED);
@@ -1179,11 +1198,18 @@ on_next_button_clicked (GtkButton *button, Totem *totem)
 static void
 on_got_redirect (BaconVideoWidget *bvw, const char *mrl, Totem *totem)
 {
-	g_message ("on_got_redirect %s", mrl);
-	//FIXME we need to check for relative paths here
+	gchar *old_mrl, *new_mrl;
+
+	old_mrl = totem_playlist_get_current_mrl (TOTEM_PLAYLIST (totem->playlist));
+	new_mrl = totem_resolve_relative_link (old_mrl, mrl);
+	g_free (old_mrl);
+
 	bacon_video_widget_close (totem->bvw);
-	bacon_video_widget_open (totem->bvw, mrl, NULL);
+	totem_gdk_window_set_waiting_cursor (totem->win->window);
+	bacon_video_widget_open (totem->bvw, new_mrl, NULL);
+	gdk_window_set_cursor (totem->win->window, NULL);
 	bacon_video_widget_play (bvw, NULL);
+	g_free (new_mrl);
 }
 
 /* This is only called when we are playing a DVD */
@@ -1239,10 +1265,9 @@ on_got_metadata_event (BaconVideoWidget *bvw, Totem *totem)
 		totem_playlist_set_title
 			(TOTEM_PLAYLIST (totem->playlist), name);
 		g_free (name);
-	} else {
-		on_playlist_change_name 	 
-			(TOTEM_PLAYLIST (totem->playlist), totem);
 	}
+	on_playlist_change_name 	 
+		(TOTEM_PLAYLIST (totem->playlist), totem);
 }
 
 static void
@@ -1385,11 +1410,16 @@ seek_slider_pressed_cb (GtkWidget *widget, GdkEventButton *event, Totem *totem)
 {
 	totem->seek_lock = TRUE;
 	totem_statusbar_set_seeking (TOTEM_STATUSBAR (totem->statusbar), TRUE);
+	totem_time_label_set_seeking (TOTEM_TIME_LABEL (totem->tcw_time_label), TRUE);
 
 	if (bacon_video_widget_can_direct_seek (totem->bvw) != FALSE)
 	{
-		totem->was_playing = bacon_video_widget_is_playing (totem->bvw);
-		bacon_video_widget_pause (totem->bvw);
+		if (!totem->seek_in_progress) {
+			totem->was_playing = bacon_video_widget_is_playing (totem->bvw);
+			bacon_video_widget_pause (totem->bvw);
+			/* We need to remember that we are already paused and seeking */
+			totem->seek_in_progress = TRUE;
+		}
 	}
 
 	return FALSE;
@@ -1408,6 +1438,9 @@ seek_slider_changed_cb (GtkAdjustment *adj, Totem *totem)
 	time = bacon_video_widget_get_stream_length (totem->bvw);
 	totem_statusbar_set_time_and_length (TOTEM_STATUSBAR (totem->statusbar),
 			(int) (pos * time / 1000), time / 1000);
+	totem_time_label_set_time
+			(TOTEM_TIME_LABEL (totem->tcw_time_label),
+			 (int) (pos * time), time);
 
 	if (bacon_video_widget_can_direct_seek (totem->bvw) != FALSE)
 		totem_action_seek (totem, pos);
@@ -1434,11 +1467,14 @@ seek_slider_released_cb (GtkWidget *widget, GdkEventButton *event, Totem *totem)
 	{
 		if (totem->was_playing)
 			bacon_video_widget_play (totem->bvw, NULL);
+		totem->seek_in_progress = FALSE;
 	} else {
 		totem_action_seek (totem, val / 65535.0);
 	}
 
 	totem_statusbar_set_seeking (TOTEM_STATUSBAR (totem->statusbar), FALSE);
+	totem_time_label_set_seeking (TOTEM_TIME_LABEL (totem->tcw_time_label),
+			FALSE);
 	return FALSE;
 }
 
@@ -1795,6 +1831,7 @@ show_controls (Totem *totem, gboolean was_fullscreen)
 			width =	widget->allocation.width;
 		}
 
+		gtk_widget_set_sensitive (menubar, TRUE);
 		gtk_widget_show (menubar);
 		gtk_widget_show (controlbar);
 		gtk_widget_show (statusbar);
@@ -1819,7 +1856,10 @@ show_controls (Totem *totem, gboolean was_fullscreen)
 			height = widget->allocation.height;
 		}
 
+		/* Hide and make the menubar unsensitive */
+		gtk_widget_set_sensitive (menubar, FALSE);
 		gtk_widget_hide (menubar);
+
 		gtk_widget_hide (controlbar);
 		gtk_widget_hide (statusbar);
 		gtk_widget_hide (totem->sidebar);
@@ -1961,6 +2001,9 @@ on_take_screenshot1_activate (GtkButton *button, Totem *totem)
 	char *filename;
 	GError *err = NULL;
 
+	if (bacon_video_widget_get_logo_mode (totem->bvw) != FALSE)
+		return;
+
 	if (bacon_video_widget_can_get_frames (totem->bvw, &err) == FALSE)
 	{
 		if (err == NULL)
@@ -2111,8 +2154,7 @@ on_skip_to1_activate (GtkButton *button, Totem *totem)
 		return;
 	}
 
-	filename = g_build_filename (DATADIR,
-			"totem", "skip_to.glade", NULL);
+	filename = totem_interface_get_full_path ("skip_to.glade");
 	totem->skipto = TOTEM_SKIPTO (totem_skipto_new (filename));
 	g_free (filename);
 
@@ -2453,7 +2495,7 @@ popup_hide (Totem *totem)
 
 	popup_timeout_remove (totem);
 
-	bacon_video_widget_set_show_cursor (totem->bvw, FALSE);
+	totem_action_set_cursor (totem, FALSE);
 
 	return FALSE;
 }
@@ -2489,7 +2531,7 @@ on_video_motion_notify_event (GtkWidget *widget, GdkEventMotion *event,
 
 	gtk_widget_show_all (totem->exit_popup);
 	gtk_widget_show_all (totem->control_popup);
-	bacon_video_widget_set_show_cursor (totem->bvw, TRUE);
+	totem_action_set_cursor (totem, TRUE);
 
 	popup_timeout_add (totem);
 	totem->popup_in_progress = FALSE;
@@ -2538,7 +2580,29 @@ on_eos_event (GtkWidget *widget, Totem *totem)
 }
 
 static gboolean
-totem_action_handle_key (Totem *totem, GdkEventKey *event)
+totem_action_handle_key_release (Totem *totem, GdkEventKey *event)
+{
+	gboolean retval = TRUE;
+
+	/* Alphabetical */
+	switch (event->keyval) {
+	case GDK_Left:
+	case GDK_Right:
+		totem_statusbar_set_seeking (TOTEM_STATUSBAR (totem->statusbar), FALSE);
+		if (bacon_video_widget_can_direct_seek (totem->bvw)) {
+			if (totem->was_playing) {
+				bacon_video_widget_play (totem->bvw, NULL);
+			}
+			totem->seek_in_progress = FALSE;
+		}
+		break;
+	}
+
+	return retval;
+}
+
+static gboolean
+totem_action_handle_key_press (Totem *totem, GdkEventKey *event)
 {
 	gboolean retval = TRUE;
 
@@ -2601,9 +2665,6 @@ totem_action_handle_key (Totem *totem, GdkEventKey *event)
 	case GDK_m:
 		bacon_video_widget_dvd_event (totem->bvw, BVW_DVD_ROOT_MENU);
 		break;
-	case GDK_Menu:
-		totem_action_menu_popup (totem, 0);
-		break;
 #ifdef HAVE_XFREE
 	case XF86XK_AudioNext:
 #endif /* HAVE_XFREE */
@@ -2634,7 +2695,11 @@ totem_action_handle_key (Totem *totem, GdkEventKey *event)
 		break;
 	case GDK_s:
 	case GDK_S:
-		on_skip_to1_activate (NULL, totem);
+		if (event->state & GDK_CONTROL_MASK) {
+			on_take_screenshot1_activate (NULL, totem);
+		} else {
+			on_skip_to1_activate (NULL, totem);
+		}
 		break;
 	case GDK_t:
 	case GDK_T:
@@ -2644,8 +2709,16 @@ totem_action_handle_key (Totem *totem, GdkEventKey *event)
 		totem_action_fullscreen (totem, FALSE);
 		break;
 	case GDK_Left:
-		if (event->state & GDK_SHIFT_MASK)
-		{
+		totem_statusbar_set_seeking (TOTEM_STATUSBAR (totem->statusbar), TRUE);
+		if (bacon_video_widget_can_direct_seek (totem->bvw)) {
+			if (!totem->seek_in_progress) {
+				totem->was_playing = bacon_video_widget_is_playing (totem->bvw);
+				bacon_video_widget_pause (totem->bvw);
+				/* We need to remember that we are already paused and seeking */
+				totem->seek_in_progress = TRUE;
+			}
+		}
+		if (event->state & GDK_SHIFT_MASK) {
 			totem_action_seek_relative (totem,
 					SEEK_BACKWARD_SHORT_OFFSET);
 		} else if (event->state & GDK_CONTROL_MASK) {
@@ -2657,19 +2730,28 @@ totem_action_handle_key (Totem *totem, GdkEventKey *event)
 		}
 		break;
 	case GDK_Right:
-		if (event->state & GDK_SHIFT_MASK)
-		{
+		totem_statusbar_set_seeking (TOTEM_STATUSBAR (totem->statusbar), TRUE);
+		if (bacon_video_widget_can_direct_seek (totem->bvw)) {
+			if (!totem->seek_in_progress) {
+				totem->was_playing = bacon_video_widget_is_playing (totem->bvw);
+				bacon_video_widget_pause (totem->bvw);
+				/* We need to remember that we are already paused and seeking */
+				totem->seek_in_progress = TRUE;
+			}
+		}
+		if (event->state & GDK_SHIFT_MASK) {
 			totem_action_seek_relative (totem,
 					SEEK_FORWARD_SHORT_OFFSET);
 		} else if (event->state & GDK_CONTROL_MASK) {
 			totem_action_seek_relative (totem,
 					SEEK_FORWARD_LONG_OFFSET);
 		} else {
-			totem_action_seek_relative (totem, SEEK_FORWARD_OFFSET);
+			totem_action_seek_relative (totem,
+					SEEK_FORWARD_OFFSET);
 		}
 		break;
 	case GDK_space:
-		if (totem_is_fullscreen (totem) != FALSE)
+		if (totem_is_fullscreen (totem) != FALSE || gtk_widget_is_focus (GTK_WIDGET (totem->bvw)) != FALSE)
 			totem_action_play_pause (totem);
 		else
 			retval = FALSE;
@@ -2681,6 +2763,12 @@ totem_action_handle_key (Totem *totem, GdkEventKey *event)
 		totem_action_volume_relative (totem, VOLUME_DOWN_OFFSET);
 		break;
 	case GDK_0:
+		if (event->state & GDK_CONTROL_MASK) {
+			totem_action_zoom_reset (totem);
+		} else {
+			totem_action_set_scale_ratio (totem, 0.5);
+		}
+		break;
 	case GDK_onehalf:
 		totem_action_set_scale_ratio (totem, 0.5);
 		break;
@@ -2690,11 +2778,35 @@ totem_action_handle_key (Totem *totem, GdkEventKey *event)
 	case GDK_2:
 		totem_action_set_scale_ratio (totem, 2);
 		break;
+	case GDK_Menu:
 	case GDK_F10:
-		if (event->state & GDK_SHIFT_MASK)
-			totem_action_menu_popup (totem, 0);
-		else
-			retval = FALSE;
+		/* Chain up if the playlist has the focuse */
+		{
+			GtkWidget *focused;
+
+			if (event->keyval != GDK_Menu && !(event->state & GDK_SHIFT_MASK))
+				return FALSE;
+
+			focused = gtk_window_get_focus (GTK_WINDOW (totem->win));
+			if (focused != NULL && gtk_widget_is_ancestor (focused, GTK_WIDGET (totem->playlist)) != FALSE)
+				return FALSE;
+		}
+
+		totem_action_menu_popup (totem, 0);
+		break;
+	case GDK_plus:
+	case GDK_KP_Add:
+		if (!(event->state & GDK_CONTROL_MASK))
+			return FALSE;
+
+		totem_action_zoom_relative (totem, ZOOM_IN_OFFSET);
+		break;
+	case GDK_minus:
+	case GDK_KP_Subtract:
+		if (!(event->state & GDK_CONTROL_MASK))
+			return FALSE;
+
+		totem_action_zoom_relative (totem, ZOOM_OUT_OFFSET);
 		break;
 	default:
 		retval = FALSE;
@@ -2762,9 +2874,21 @@ on_window_key_press_event (GtkWidget *win, GdkEventKey *event, Totem *totem)
 		case GDK_o:
 		case GDK_L:
 		case GDK_l:
+		case GDK_S:
+		case GDK_s:
 		case GDK_Right:
 		case GDK_Left:
-			return totem_action_handle_key (totem, event);
+		case GDK_plus:
+		case GDK_KP_Add:
+		case GDK_minus:
+		case GDK_KP_Subtract:
+		case GDK_0:
+			if (event->type == GDK_KEY_PRESS) {
+				return totem_action_handle_key_press (totem, event);
+			}
+			else {
+				return totem_action_handle_key_release (totem, event);
+			}
 	}
 
 	/* If we have modifiers, and either Ctrl, Mod1 (Alt), or any
@@ -2778,7 +2902,12 @@ on_window_key_press_event (GtkWidget *win, GdkEventKey *event, Totem *totem)
 			|| (event->state & GDK_MOD5_MASK)))
 		return FALSE;
 
-	return totem_action_handle_key (totem, event);
+	if (event->type == GDK_KEY_PRESS) {
+		return totem_action_handle_key_press (totem, event);
+	}
+	else {
+		return totem_action_handle_key_release (totem, event);
+	}
 }
 
 static int
@@ -3113,9 +3242,11 @@ totem_callback_connect (Totem *totem)
 			G_CALLBACK (vol_slider_released_cb), totem);
 
 	/* Connect the keys */
-	gtk_widget_add_events (totem->win, GDK_KEY_PRESS_MASK);
+	gtk_widget_add_events (totem->win, GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK);
 	g_signal_connect (G_OBJECT(totem->win), "key_press_event",
 			G_CALLBACK (on_window_key_press_event), totem);
+	g_signal_connect (G_OBJECT(totem->win), "key_release_event",
+			G_CALLBACK (on_window_key_press_event), totem);
 
 	/* Connect the mouse wheel */
 	gtk_widget_add_events (totem->win, GDK_SCROLL_MASK);
@@ -3304,9 +3435,12 @@ video_widget_create (Totem *totem)
 			GTK_WIDGET (totem->bvw));
 
 	/* Events for the widget video window as well */
-	gtk_widget_add_events (GTK_WIDGET (totem->bvw), GDK_KEY_PRESS_MASK);
+	gtk_widget_add_events (GTK_WIDGET (totem->bvw),
+			GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK);
 	g_signal_connect (G_OBJECT(totem->bvw), "key_press_event",
 			G_CALLBACK (on_window_key_press_event), totem);
+	g_signal_connect (G_OBJECT(totem->bvw), "key_release_event",
+			G_CALLBACK (on_window_key_press_event), totem);
 
 	g_signal_connect (G_OBJECT (totem->bvw), "drag_data_received",
 			G_CALLBACK (drop_video_cb), totem);
@@ -3432,6 +3566,7 @@ main (int argc, char **argv)
 	/* Init totem itself */
 	totem->prev_volume = -1;
 	totem->gc = gc;
+	totem->cursor_shown = TRUE;
 
 	/* Main window */
 	totem->xml = totem_interface_load ("totem.glade", _("main window"),
@@ -3498,6 +3633,7 @@ main (int argc, char **argv)
 	/* Show ! gtk_main_iteration trickery to show all the widgets
 	 * we have so far */
 	gtk_widget_show (totem->win);
+	totem_gdk_window_set_waiting_cursor (totem->win->window);
 	update_fullscreen_size (totem);
 	long_action ();
 
@@ -3519,6 +3655,7 @@ main (int argc, char **argv)
 	} else {
 		totem_action_set_mrl (totem, NULL);
 	}
+	gdk_window_set_cursor (totem->win->window, NULL);
 
 	if (bacon_message_connection_get_is_server (totem->conn) != FALSE)
 	{
